import logging
from collections import defaultdict
from typing import List, Dict

from twisted.internet.defer import DeferredList, inlineCallbacks, Deferred

from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.client.controller.SearchObjectCacheController import \
    SearchObjectCacheController
from peek_core_search._private.tuples.search_object.SearchObjectUpdateDateTuple import \
    SearchObjectUpdateDateTuple
from vortex.DeferUtil import vortexLogFailure
from vortex.Payload import Payload
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope
from vortex.VortexABC import SendVortexMsgResponseCallable
from vortex.VortexFactory import VortexFactory

logger = logging.getLogger(__name__)

clientSearchObjectWatchUpdateFromDeviceFilt = {
    'key': "clientSearchObjectWatchUpdateFromDevice"
}
clientSearchObjectWatchUpdateFromDeviceFilt.update(searchFilt)


# ModelSet HANDLER
class SearchObjectCacheHandler(object):
    def __init__(self, cacheController: SearchObjectCacheController,
                 clientId: str):
        """ App SearchObject Handler

        This class handles the custom needs of the desktop/mobile apps observing searchObjects.

        """
        self._cacheController = cacheController
        self._clientId = clientId

        self._epObserve = PayloadEndpoint(
            clientSearchObjectWatchUpdateFromDeviceFilt, self._processObserve
        )

        self._uuidsObserving = set()

    def shutdown(self):
        self._epObserve.shutdown()
        self._epObserve = None

    # ---------------
    # Process update from the server

    def notifyOfSearchObjectUpdate(self, chunkKeys: List[str]):
        """ Notify of SearchObject Updates

        This method is called by the client.SearchObjectCacheController when it receives updates
        from the server.

        """
        vortexUuids = set(VortexFactory.getRemoteVortexUuids()) & self._uuidsObserving
        self._uuidsObserving = vortexUuids

        payloadsByVortexUuid = defaultdict(Payload)

        for chunkKey in chunkKeys:
            encodedSearchObjectChunk = self._cacheController.searchObject(chunkKey)

            # Queue up the required client notifications
            for vortexUuid in vortexUuids:
                logger.debug("Sending unsolicited searchObject %s to vortex %s",
                             chunkKey, vortexUuid)
                payloadsByVortexUuid[vortexUuid].tuples.append(encodedSearchObjectChunk)

        # Send the updates to the clients
        dl = []
        for vortexUuid, payload in list(payloadsByVortexUuid.items()):
            payload.filt = clientSearchObjectWatchUpdateFromDeviceFilt

            # Serialise in thread, and then send.
            d = payload.makePayloadEnvelopeDefer()
            d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
            d.addCallback(VortexFactory.sendVortexMsg, destVortexUuid=vortexUuid)
            dl.append(d)

        # Log the errors, otherwise we don't care about them
        dl = DeferredList(dl, fireOnOneErrback=True)
        dl.addErrback(vortexLogFailure, logger, consumeError=True)

    # ---------------
    # Process observes from the devices
    @inlineCallbacks
    def _processObserve(self, payloadEnvelope: PayloadEnvelope,
                        vortexUuid: str,
                        sendResponse: SendVortexMsgResponseCallable,
                        **kwargs):

        payload = yield payloadEnvelope.decodePayloadDefer()

        updateDatesTuples: SearchObjectUpdateDateTuple = payload.tuples[0]

        self._uuidsObserving.add(vortexUuid)

        yield self._replyToObserve(payload.filt,
                                   updateDatesTuples.updateDateByChunkKey,
                                   sendResponse,
                                   vortexUuid)

    # ---------------
    # Reply to device observe

    @inlineCallbacks
    def _replyToObserve(self, filt,
                        lastUpdateBySearchObjectKey: Dict[str, str],
                        sendResponse: SendVortexMsgResponseCallable,
                        vortexUuid: str) -> None:
        """ Reply to Observe

        The client has told us that it's observing a new set of searchObjects, and the lastUpdate
        it has for each of those searchObjects. We will send them the searchObjects that are out of date
        or missing.

        :param filt: The payload filter to respond to.
        :param lastUpdateBySearchObjectKey: The dict of searchObjectKey:lastUpdate
        :param sendResponse: The callable provided by the Vortex (handy)
        :returns: None

        """
        yield None

        searchObjectTuplesToSend = []

        # Check and send any updates
        for searchObjectKey, lastUpdate in lastUpdateBySearchObjectKey.items():
            if vortexUuid not in VortexFactory.getRemoteVortexUuids():
                logger.debug("Vortex %s is offline, stopping update")
                return

            searchObjectKey = int(searchObjectKey)

            # NOTE: lastUpdate can be null.
            encodedSearchObjectTuple = self._cacheController.searchObject(searchObjectKey)
            if not encodedSearchObjectTuple:
                logger.debug("SearchObject %s is not in the cache" % searchObjectKey)
                continue

            # We are king, If it's it's not our version, it's the wrong version ;-)
            logger.debug("%s, %s,  %s",
                         encodedSearchObjectTuple.lastUpdate == lastUpdate,
                         encodedSearchObjectTuple.lastUpdate, lastUpdate)

            if encodedSearchObjectTuple.lastUpdate == lastUpdate:
                logger.debug("SearchObject %s matches the cache" % searchObjectKey)
                continue

            searchObjectTuplesToSend.append(encodedSearchObjectTuple)
            logger.debug("Sending searchObject %s from the cache" % searchObjectKey)

        # Send the payload to the frontend
        payload = Payload(filt=filt, tuples=searchObjectTuplesToSend)
        d: Deferred = payload.makePayloadEnvelopeDefer()
        d.addCallback(lambda payloadEnvelope: payloadEnvelope.toVortexMsgDefer())
        d.addCallback(sendResponse)
        d.addErrback(vortexLogFailure, logger, consumeError=True)



import logging

from twisted.internet.defer import Deferred

from peek_core_search._private.tuples.AdminStatusTuple import AdminStatusTuple
from vortex.TupleAction import TupleActionABC
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC
from vortex.handler.TupleDataObservableHandler import TupleDataObservableHandler

logger = logging.getLogger(__name__)


class StatusController:
    def __init__(self):
        self._status = AdminStatusTuple()
        self._tupleObservable = None

    def setTupleObservable(self, tupleObserver: TupleDataObservableHandler):
        self._tupleObserver = tupleObserver

    def shutdown(self):
        self._tupleObserver = None

    @property
    def status(self):
        return self._status


    # ---------------
    # Search Index Compiler Methods

    def setSearchIndexCompilerStatus(self, state: bool, queueSize: int):
        self._status.searchIndexCompilerQueueStatus = state
        self._status.searchIndexCompilerQueueSize = queueSize
        self._notify()

    def addToSearchIndexCompilerTotal(self, delta: int):
        self._status.searchIndexCompilerQueueProcessedTotal += delta
        self._notify()

    def setSearchIndexCompilerError(self, error: str):
        self._status.searchIndexCompilerQueueLastError = error
        self._notify()


    # ---------------
    # Search Object Compiler Methods

    def setSearchObjectCompilerStatus(self, state: bool, queueSize: int):
        self._status.searchObjectCompilerQueueStatus = state
        self._status.searchObjectCompilerQueueSize = queueSize
        self._notify()

    def addToSearchObjectCompilerTotal(self, delta: int):
        self._status.searchObjectCompilerQueueProcessedTotal += delta
        self._notify()

    def setSearchObjectCompilerError(self, error: str):
        self._status.searchObjectCompilerQueueLastError = error
        self._notify()

    # ---------------
    # Notify Methods

    def _notify(self):
        self._tupleObserver.notifyOfTupleUpdate(
            TupleSelector(AdminStatusTuple.tupleType(), {})
        )



import logging
from typing import Dict, List

from twisted.internet.defer import inlineCallbacks
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.server.client_handlers.ClientChunkLoadRpc import \
    ClientChunkLoadRpc
from peek_core_search._private.storage.EncodedSearchIndexChunk import \
    EncodedSearchIndexChunk

logger = logging.getLogger(__name__)

clientSearchIndexUpdateFromServerFilt = dict(key="clientSearchIndexUpdateFromServer")
clientSearchIndexUpdateFromServerFilt.update(searchFilt)


class SearchIndexCacheController:
    """ SearchIndex Cache Controller

    The SearchIndex cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    LOAD_CHUNK = 32

    def __init__(self, clientId: str):
        self._clientId = clientId
        self._webAppHandler = None
        self._fastKeywordController = None

        #: This stores the cache of searchIndex data for the clients
        self._cache: Dict[int, EncodedSearchIndexChunk] = {}

        self._endpoint = PayloadEndpoint(clientSearchIndexUpdateFromServerFilt,
                                         self._processSearchIndexPayload)

    def setSearchIndexCacheHandler(self, handler):
        self._webAppHandler = handler

    def setFastKeywordController(self, fastKeywordController):
        self._fastKeywordController = fastKeywordController

    @inlineCallbacks
    def start(self):
        yield self.reloadCache()

    def shutdown(self):
        self._webAppHandler = None
        self._fastKeywordController = None
        self._endpoint.shutdown()
        self._endpoint = None

        self._cache = {}

    @inlineCallbacks
    def reloadCache(self):
        self._cache = {}

        offset = 0
        while True:
            logger.info(
                "Loading SearchIndexChunk %s to %s" % (offset, offset + self.LOAD_CHUNK))
            encodedChunkTuples: List[EncodedSearchIndexChunk] = (
                yield ClientChunkLoadRpc.loadSearchIndexChunks(offset, self.LOAD_CHUNK)
            )

            if not encodedChunkTuples:
                break

            self._loadSearchIndexIntoCache(encodedChunkTuples)

            offset += self.LOAD_CHUNK

    @inlineCallbacks
    def _processSearchIndexPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        paylod = yield payloadEnvelope.decodePayloadDefer()
        searchIndexTuples: List[EncodedSearchIndexChunk] = paylod.tuples
        self._loadSearchIndexIntoCache(searchIndexTuples)

    def _loadSearchIndexIntoCache(self,
                                  encodedChunkTuples: List[EncodedSearchIndexChunk]):
        chunkKeysUpdated: List[str] = []

        for t in encodedChunkTuples:

            if (not t.chunkKey in self._cache or
                    self._cache[t.chunkKey].lastUpdate != t.lastUpdate):
                self._cache[t.chunkKey] = t
                chunkKeysUpdated.append(t.chunkKey)

        logger.debug("Received searchIndex updates from server, %s", chunkKeysUpdated)

        self._webAppHandler.notifyOfSearchIndexUpdate(chunkKeysUpdated)
        self._fastKeywordController.notifyOfSearchIndexUpdate(chunkKeysUpdated)

    def searchIndex(self, chunkKey) -> EncodedSearchIndexChunk:
        return self._cache.get(chunkKey)

    def searchIndexKeys(self) -> List[int]:
        return list(self._cache)



import logging

from sqlalchemy import Column
from sqlalchemy import Integer, String

from peek_core_search._private.PluginNames import searchTuplePrefix
from vortex.Tuple import Tuple, addTupleType
from .DeclarativeBase import DeclarativeBase

logger = logging.getLogger(__name__)


@addTupleType
class SearchObjectCompilerQueue(Tuple, DeclarativeBase):
    __tablename__ = 'SearchObjectCompilerQueue'
    __tupleType__ = searchTuplePrefix + 'SearchObjectCompilerQueueTable'

    id = Column(Integer, primary_key=True, autoincrement=True)
    chunkKey = Column(Integer, primary_key=True)





import logging
from collections import defaultdict
from typing import Dict, List, Optional

import ujson
from twisted.internet.defer import inlineCallbacks
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload

from peek_core_search._private.PluginNames import searchFilt
from peek_core_search._private.server.client_handlers.ClientChunkLoadRpc import \
    ClientChunkLoadRpc
from peek_core_search._private.storage.EncodedSearchObjectChunk import \
    EncodedSearchObjectChunk
from vortex.PayloadEndpoint import PayloadEndpoint
from vortex.PayloadEnvelope import PayloadEnvelope

from peek_core_search._private.storage.SearchObjectTypeTuple import SearchObjectTypeTuple
from peek_core_search._private.tuples.search_object.SearchResultObjectRouteTuple import \
    SearchResultObjectRouteTuple
from peek_core_search._private.tuples.search_object.SearchResultObjectTuple import \
    SearchResultObjectTuple
from peek_core_search._private.worker.tasks._CalcChunkKey import makeSearchObjectChunkKey

logger = logging.getLogger(__name__)

clientSearchObjectUpdateFromServerFilt = dict(key="clientSearchObjectUpdateFromServer")
clientSearchObjectUpdateFromServerFilt.update(searchFilt)


class SearchObjectCacheController:
    """ SearchObject Cache Controller

    The SearchObject cache controller stores all the chunks in memory,
    allowing fast access from the mobile and desktop devices.

    """

    LOAD_CHUNK = 32

    def __init__(self, clientId: str):
        self._clientId = clientId
        self._webAppHandler = None

        #: This stores the cache of searchObject data for the clients
        self._cache: Dict[int, EncodedSearchObjectChunk] = {}

        self._endpoint = PayloadEndpoint(clientSearchObjectUpdateFromServerFilt,
                                         self._processSearchObjectPayload)

    def setSearchObjectCacheHandler(self, handler):
        self._webAppHandler = handler

    @inlineCallbacks
    def start(self):
        yield self.reloadCache()

    def shutdown(self):
        self._tupleObservable = None

        self._endpoint.shutdown()
        self._endpoint = None

        self._cache = {}

    @inlineCallbacks
    def reloadCache(self):
        self._cache = {}

        offset = 0
        while True:
            logger.info(
                "Loading SearchObjectChunk %s to %s" % (offset, offset + self.LOAD_CHUNK))
            encodedChunkTuples: List[EncodedSearchObjectChunk] = (
                yield ClientChunkLoadRpc.loadSearchObjectChunks(offset, self.LOAD_CHUNK)
            )

            if not encodedChunkTuples:
                break

            self._loadSearchObjectIntoCache(encodedChunkTuples)

            offset += self.LOAD_CHUNK

    @inlineCallbacks
    def _processSearchObjectPayload(self, payloadEnvelope: PayloadEnvelope, **kwargs):
        paylod = yield payloadEnvelope.decodePayloadDefer()
        searchObjectTuples: List[EncodedSearchObjectChunk] = paylod.tuples
        self._loadSearchObjectIntoCache(searchObjectTuples)

    def _loadSearchObjectIntoCache(self,
                                   encodedChunkTuples: List[EncodedSearchObjectChunk]):
        chunkKeysUpdated: List[str] = []

        for t in encodedChunkTuples:

            if (not t.chunkKey in self._cache or
                    self._cache[t.chunkKey].lastUpdate != t.lastUpdate):
                self._cache[t.chunkKey] = t
                chunkKeysUpdated.append(t.chunkKey)

        logger.debug("Received searchObject updates from server, %s", chunkKeysUpdated)

        self._webAppHandler.notifyOfSearchObjectUpdate(chunkKeysUpdated)

    def searchObject(self, chunkKey) -> EncodedSearchObjectChunk:
        return self._cache.get(chunkKey)

    def searchObjectKeys(self) -> List[int]:
        return list(self._cache)

    @deferToThreadWrapWithLogger(logger)
    def getObjects(self, objectTypeId: Optional[int],
                   objectIds: List[int]) -> List[SearchResultObjectTuple]:
        return self.getObjectsBlocking(objectTypeId, objectIds)

    def getObjectsBlocking(self, objectTypeId: Optional[int],
                           objectIds: List[int]) -> List[SearchResultObjectTuple]:

        objectIdsByChunkKey = defaultdict(list)
        for objectId in objectIds:
            objectIdsByChunkKey[makeSearchObjectChunkKey(objectId)].append(objectId)

        foundObjects: List[SearchResultObjectTuple] = []
        for chunkKey, subObjectIds in objectIdsByChunkKey.items():
            foundObjects += self._getObjectsForChunkBlocking(
                chunkKey, objectTypeId, subObjectIds
            )

        return foundObjects

    def _getObjectsForChunkBlocking(self, chunkKey: str,
                                    objectTypeId: Optional[int],
                                    objectIds: List[int]
                                    ) -> List[SearchResultObjectTuple]:

        chunk = self.searchObject(chunkKey)
        if not chunk:
            return []

        objectPropsByIdStr = Payload().fromEncodedPayload(chunk.encodedData).tuples[0]
        objectPropsById = ujson.loads(objectPropsByIdStr)

        foundObjects: List[SearchResultObjectTuple] = []

        for objectId in objectIds:
            if str(objectId) not in objectPropsById:
                logger.warning(
                    "Search object id %s is missing from index, chunkKey %s",
                    objectId, chunkKey
                )
                continue

            # Reconstruct the data
            objectProps: {} = ujson.loads(objectPropsById[str(objectId)])

            # Get out the object type
            thisObjectTypeId = objectProps['_otid_']
            del objectProps['_otid_']

            # If the property is set, then make sure it matches
            if objectTypeId is not None and objectTypeId != thisObjectTypeId:
                continue

            # Get out the routes
            routes: List[List[str]] = objectProps['_r_']
            del objectProps['_r_']

            # Get the key
            objectKey: str = objectProps['key']

            # Create the new object
            newObject = SearchResultObjectTuple()
            foundObjects.append(newObject)

            newObject.id = objectId
            newObject.key = objectKey
            newObject.objectType = SearchObjectTypeTuple(id=thisObjectTypeId)
            newObject.properties = objectProps

            for route in routes:
                newRoute = SearchResultObjectRouteTuple()
                newObject.routes.append(newRoute)

                newRoute.title = route[0]
                newRoute.path = route[1]

        return foundObjects



import json
import logging
from collections import defaultdict
from typing import Union, List, Optional, Dict

from twisted.internet.defer import Deferred

from peek_core_search._private.client.controller.SearchIndexCacheController import \
    SearchIndexCacheController
from peek_core_search._private.client.controller.SearchObjectCacheController import \
    SearchObjectCacheController
from peek_core_search._private.storage.EncodedSearchIndexChunk import \
    EncodedSearchIndexChunk
from peek_core_search._private.storage.EncodedSearchObjectChunk import \
    EncodedSearchObjectChunk
from peek_core_search._private.storage.SearchObjectTypeTuple import \
    SearchObjectTypeTuple
from peek_core_search._private.tuples.search_object.SearchResultObjectRouteTuple import \
    SearchResultObjectRouteTuple
from peek_core_search._private.tuples.search_object.SearchResultObjectTuple import \
    SearchResultObjectTuple
from peek_core_search._private.worker.tasks._CalcChunkKey import makeSearchIndexChunkKey, \
    makeSearchObjectChunkKey
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleSelector import TupleSelector
from vortex.handler.TupleDataObservableHandler import TuplesProviderABC

logger = logging.getLogger(__name__)


class ClientSearchObjectResultTupleProvider(TuplesProviderABC):
    def __init__(self, searchIndexCacheHandler: SearchIndexCacheController,
                 searchObjectCacheHandler: SearchObjectCacheController):
        self._searchIndexCacheHandler = searchIndexCacheHandler
        self._searchObjectCacheHandler = searchObjectCacheHandler

    @deferToThreadWrapWithLogger(logger)
    def makeVortexMsg(self, filt: dict,
                      tupleSelector: TupleSelector) -> Union[Deferred, bytes]:
        propertyName: Optional[str] = tupleSelector.selector["propertyName"]
        objectTypeId: Optional[int] = tupleSelector.selector["objectTypeId"]
        keywords: List[str] = tupleSelector.selector["keywords"]

        # GET THE OBJECT IDS FROM KEYWORD
        keysByChunkKey = defaultdict(list)
        for keyword in keywords:
            keysByChunkKey[makeSearchIndexChunkKey(keyword)].append(keyword)

        foundObjectIdCounts: Dict[int, int] = defaultdict(lambda: 0)
        for chunkKey, subKeys in keysByChunkKey.items():
            encodedChunk = self._searchIndexCacheHandler.searchIndex(chunkKey)
            if encodedChunk:
                for objId in self._getObjectIds(encodedChunk, propertyName, subKeys):
                    foundObjectIdCounts[objId] += 1

        # Return all the object IDs that have the most keyword matches
        foundObjectIds: List[int] = []
        maxCount = len(keywords)

        for objectId, maxObjectNum in foundObjectIdCounts.items():
            if maxObjectNum == maxCount:
                foundObjectIds.append(objectId)

        # LIMIT TO 20
        foundObjectIds = foundObjectIds[:20]

        # GET OBJECTS
        foundObjects = self._searchObjectCacheHandler \
            .getObjectsBlocking(objectTypeId, foundObjectIds)

        # Create the vortex message
        return Payload(filt, tuples=foundObjects).makePayloadEnvelope().toVortexMsg()

    def _getObjectIds(self, chunk: EncodedSearchIndexChunk,
                      propertyName: Optional[str],
                      keywords: List[str]) -> List[int]:

        chunkData = Payload().fromEncodedPayload(chunk.encodedData).tuples

        indexByKeyword = {item[0]: item for item in chunkData}
        foundObjectIds: List[int] = []

        for keyword in keywords:
            if keyword not in indexByKeyword:
                logger.warning(
                    "Search keyword %s is missing from index, chunkKey %s",
                    keyword, chunk.chunkKey
                )
                continue

            keywordIndex = indexByKeyword[keyword]

            # If the property is set, then make sure it matches
            if propertyName is not None and keywordIndex[1] != propertyName:
                continue

            # This is stored as a string, so we don't have to construct
            # so much data when deserialising the chunk
            foundObjectIds += json.loads(keywordIndex[2])

        return foundObjectIds



from sqlalchemy import Column, LargeBinary
from sqlalchemy import Integer, String
from sqlalchemy.sql.schema import Index

from peek_core_search._private.PluginNames import searchTuplePrefix
from peek_core_search._private.storage.DeclarativeBase import DeclarativeBase
from vortex.Tuple import Tuple, addTupleType


@addTupleType
class EncodedSearchIndexChunk(Tuple, DeclarativeBase):
    __tablename__ = 'EncodedSearchIndexChunk'
    __tupleType__ = searchTuplePrefix + 'EncodedSearchIndexChunkTable'

    ENCODED_DATA_KEYWORD_NUM = 0
    ENCODED_DATA_PROPERTY_MAME_NUM = 1
    ENCODED_DATA_OBJECT_IDS_JSON_INDEX = 2

    id = Column(Integer, primary_key=True, autoincrement=True)

    chunkKey = Column(Integer, primary_key=True)
    encodedData = Column(LargeBinary, nullable=False)
    encodedHash = Column(String, nullable=False)
    lastUpdate = Column(String, nullable=False)

    __table_args__ = (
        Index("idx_EncodedSearchIndex_chunkKey", chunkKey, unique=True),
    )



""" Fast Graph DB

This module stores a memory resident model of a graph network.

"""
import logging
from collections import defaultdict
from datetime import datetime
from typing import Optional, List, Dict, Tuple

import pytz
import ujson
from twisted.internet.defer import inlineCallbacks, Deferred
from vortex.DeferUtil import deferToThreadWrapWithLogger
from vortex.Payload import Payload
from vortex.TupleAction import TupleActionABC
from vortex.handler.TupleActionProcessor import TupleActionProcessorDelegateABC

from peek_core_search._private.client.controller.SearchIndexCacheController import \
    SearchIndexCacheController
from peek_core_search._private.client.controller.SearchObjectCacheController import \
    SearchObjectCacheController
from peek_core_search._private.storage.EncodedSearchIndexChunk import \
    EncodedSearchIndexChunk
from peek_core_search._private.tuples.KeywordAutoCompleteTupleAction import \
    KeywordAutoCompleteTupleAction
from peek_core_search._private.worker.tasks.ImportSearchIndexTask import _splitKeywords

logger = logging.getLogger(__name__)


class FastKeywordController(TupleActionProcessorDelegateABC):
    def __init__(self, objectCacheController: SearchObjectCacheController,
                 indexCacheController: SearchIndexCacheController):
        self._objectCacheController = objectCacheController
        self._indexCacheController = indexCacheController
        self._keywordsByPropertyKeyByChunkKey: Dict[str, Dict[str, List[str]]] = \
            defaultdict(lambda: defaultdict(list))

    def shutdown(self):
        self._objectCacheController = None
        self._indexCacheController = None
        self._keywordsByPropertyKeyByChunkKey = {}

    @inlineCallbacks
    def processTupleAction(self, tupleAction: TupleActionABC) -> Deferred:
        assert isinstance(tupleAction, KeywordAutoCompleteTupleAction), \
            "Tuple is not a KeywordAutoCompleteTupleAction"

        startTime = datetime.now(pytz.utc)

        objectIds = yield self.getObjectIdsForSearchString(
            tupleAction.searchString, tupleAction.propertyKey
        )

        results = yield self._objectCacheController.getObjects(
            tupleAction.objectTypeId, objectIds
        )

        logger.debug("Completed search for |%s|, returning %s objects, in %s",
                     tupleAction.searchString,
                     len(results), (datetime.now(pytz.utc) - startTime))

        return results

    @deferToThreadWrapWithLogger(logger)
    def getObjectIdsForSearchString(self, searchString: str,
                                    argPropertyKey: Optional[str]) -> Deferred:
        """ Get ObjectIds For Search String

        :rtype List[int]

        """
        splitKws = _splitKeywords(searchString)
        if not splitKws:
            return []

        results = [[] for _ in splitKws]

        for chunkData in self._keywordsByPropertyKeyByChunkKey.values():
            for (propertyKey, keyword), objectIdStr in chunkData.items():
                if argPropertyKey is not None and argPropertyKey != propertyKey:
                    continue

                for index, partialKw in enumerate(splitKws):
                    if partialKw in keyword:
                        results[index] += ujson.loads(objectIdStr)

        objectIdsUnion = set(results[0])
        for objectIds in results[1:]:
            objectIdsUnion &= set(objectIds)

        return list(objectIdsUnion)[:50]

    @inlineCallbacks
    def notifyOfSearchIndexUpdate(self, chunkKeys: List[str]):
        """ Notify of Segment Updates

        This method is called by the client.SearchIndexCacheController when it receives
         updates from the server.

        """
        for chunkKey in chunkKeys:
            encodedChunkTuple = self._indexCacheController.searchIndex(chunkKey)
            yield self._unpackKeywordsFromChunk(encodedChunkTuple)

    @deferToThreadWrapWithLogger(logger)
    def _unpackKeywordsFromChunk(self, chunk: EncodedSearchIndexChunk) -> None:

        chunkDataTuples = Payload().fromEncodedPayload(chunk.encodedData).tuples

        chunkData: Dict[Tuple[str, str], str] = {}

        for data in chunkDataTuples:
            keyword = data[EncodedSearchIndexChunk.ENCODED_DATA_KEYWORD_NUM]
            propertyName = data[EncodedSearchIndexChunk.ENCODED_DATA_PROPERTY_MAME_NUM]
            objectIdsJson = data[
                EncodedSearchIndexChunk.ENCODED_DATA_OBJECT_IDS_JSON_INDEX]
            chunkData[(propertyName, keyword)] = objectIdsJson

        self._keywordsByPropertyKeyByChunkKey[chunk.chunkKey] = chunkData
